import colorsys
import cv2
from PIL import Image
import pkg_resources
import pytesseract

from .LivescoreBase import LivescoreBase
from .details import Alliance2018, OngoingMatchDetails


class Livescore2018(LivescoreBase):
    def __init__(self, **kwargs):
        super(Livescore2018, self).__init__(2018, **kwargs)
        self._match_key = None
        self._match_name = None

        # Power up templates
        template_path = pkg_resources.resource_filename(__name__, 'templates')
        self._templates = {
            'boost': cv2.imread(template_path + '/2018_boost.png'),
            'force': cv2.imread(template_path + '/2018_force.png'),
            'levitate': cv2.imread(template_path + '/2018_levitate.png'),
        }

    def _getMatchKeyName(self, img, debug_img):
        if self._match_key is None:
            tl = self._transformPoint((159, 130))
            br = self._transformPoint((570, 162))

            config = '--psm 7'
            raw_match_name = pytesseract.image_to_string(
                Image.fromarray(self._getImgCropThresh(img, tl, br)), config=config).strip()
            self._match_key = self._getMatchKey(raw_match_name)
            if self._match_key:
                self._match_name = raw_match_name
            else:
                self._match_name = None

            if self._debug:
                box = self._cornersToBox(tl, br)
                self._drawBox(debug_img, box, (0, 255, 0))

        return self._match_key, self._match_name

    def _getTimeAndMode(self, img, debug_img):
        # Find time remaining
        tl = self._transformPoint((617, 14))
        br = self._transformPoint((665, 38))
        time_remaining = self._parseDigits(self._getImgCropThresh(img, tl, br))

        # Determine mode: 'pre_match', 'auto', 'teleop', or 'post_match'
        mode_point = self._transformPoint((497, 26))
        mode_point2 = self._transformPoint((581, 26))
        mode_sample = img[mode_point[1], mode_point[0], :]
        mode_sample2 = img[mode_point2[1], mode_point2[0], :]

        hsv = colorsys.rgb_to_hsv(float(mode_sample[2])/255, float(mode_sample[1])/255, float(mode_sample[0])/255)
        hsv2 = colorsys.rgb_to_hsv(float(mode_sample2[2])/255, float(mode_sample2[1])/255, float(mode_sample2[0])/255)

        if time_remaining is None:
            mode = None
        if time_remaining == 0:
            if hsv[1] > 0.6 and hsv2[1] > 0.6:  # Both saturated
                mode = 'post_match'
            elif hsv[1] > 0.6:  # First saturated
                mode = 'auto'  # End of auton
            else:
                mode = 'pre_match'
        elif time_remaining <= 15 and hsv2[1] < 0.6:
            mode = 'auto'
        else:
            mode = 'teleop'

        if self._debug:
            box = self._cornersToBox(tl, br)
            self._drawBox(debug_img, box, (0, 255, 0))
            cv2.circle(debug_img, mode_point, 2, (0, 255, 0), -1)
            cv2.circle(debug_img, mode_point2, 2, (0, 255, 0), -1)

        return time_remaining, mode

    def _getFlipped(self, img, debug_img):
        # Sample point to determine red/blue side
        color_point = self._transformPoint((496, 52))
        color_sample = img[color_point[1], color_point[0], :]
        is_flipped = color_sample[0] > color_sample[2]  # More blue than red

        if self._debug:
            cv2.circle(debug_img, color_point, 2, (0, 255, 0), -1)

        return is_flipped

    def _getScores(self, img, debug_img, is_flipped):
        # Left score limits
        left_tl = self._transformPoint((497, 61))
        left_br = self._transformPoint((618, 114))
        # Right score limits
        right_tl = self._transformPoint((661, 61))
        right_br = self._transformPoint((779, 114))

        left_score = self._parseDigits(self._getImgCropThresh(img, left_tl, left_br, white=True))
        right_score = self._parseDigits(self._getImgCropThresh(img, right_tl, right_br, white=True))

        if is_flipped:
            red_score = right_score
            blue_score = left_score
        else:
            red_score = left_score
            blue_score = right_score

        if self._debug:
            left_box = self._cornersToBox(left_tl, left_br)
            right_box = self._cornersToBox(right_tl, right_br)
            self._drawBox(debug_img, left_box, (255, 255, 0) if is_flipped else (255, 0, 255))
            self._drawBox(debug_img, right_box, (255, 0, 255) if is_flipped else (255, 255, 0))

        return red_score, blue_score

    def _getVaultInfo(self, img, debug_img, is_flipped):
        # Left powerups
        left_force_tl = self._transformPoint((75, 83))
        left_force_br = self._transformPoint((93, 106))
        left_force_loc = self._transformPoint((40, 84))
        left_boost_tl = self._transformPoint((145, 83))
        left_boost_br = self._transformPoint((163, 106))
        left_boost_loc = self._transformPoint((198, 84))
        left_levitate_tl = self._transformPoint((110, 65))
        left_levitate_br = self._transformPoint((127, 86))
        left_levitate_loc = self._transformPoint((99, 44))

        # Right powerups
        right_force_tl = self._transformPoint((1279 - 163, 83))
        right_force_br = self._transformPoint((1279 - 145, 106))
        right_force_loc = self._transformPoint((1279 - 198, 84))
        right_boost_tl = self._transformPoint((1279 - 93, 83))
        right_boost_br = self._transformPoint((1279 - 75, 106))
        right_boost_loc = self._transformPoint((1279 - 40, 84))
        right_levitate_tl = self._transformPoint((1279 - 128, 65))
        right_levitate_br = self._transformPoint((1279 - 110, 86))
        right_levitate_loc = self._transformPoint((1279 - 99, 44))

        # Counts
        left_boost_count = self._parseDigits(self._getImgCropThresh(img, left_boost_tl, left_boost_br))
        left_force_count = self._parseDigits(self._getImgCropThresh(img, left_force_tl, left_force_br))
        left_levitate_count = self._parseDigits(self._getImgCropThresh(img, left_levitate_tl, left_levitate_br))
        right_boost_count = self._parseDigits(self._getImgCropThresh(img, right_boost_tl, right_boost_br))
        right_force_count = self._parseDigits(self._getImgCropThresh(img, right_force_tl, right_force_br))
        right_levitate_count = self._parseDigits(self._getImgCropThresh(img, right_levitate_tl, right_levitate_br))

        # Played
        left_boost_played = self._checkSaturated(img, left_boost_loc)
        left_force_played = self._checkSaturated(img, left_force_loc)
        left_levitate_played = self._checkSaturated(img, left_levitate_loc)
        right_boost_played = self._checkSaturated(img, right_boost_loc)
        right_force_played = self._checkSaturated(img, right_force_loc)
        right_levitate_played = self._checkSaturated(img, right_levitate_loc)

        if is_flipped:
            red_boost_count = right_boost_count
            red_force_count = right_force_count
            red_levitate_count = right_levitate_count
            red_boost_played = right_boost_played
            red_force_played = right_force_played
            red_levitate_played = right_levitate_played
            blue_boost_count = left_boost_count
            blue_force_count = left_force_count
            blue_levitate_count = left_levitate_count
            blue_boost_played = left_boost_played
            blue_force_played = left_force_played
            blue_levitate_played = left_levitate_played
        else:
            red_boost_count = left_boost_count
            red_force_count = left_force_count
            red_levitate_count = left_levitate_count
            red_boost_played = left_boost_played
            red_force_played = left_force_played
            red_levitate_played = left_levitate_played
            blue_boost_count = right_boost_count
            blue_force_count = right_force_count
            blue_levitate_count = right_levitate_count
            blue_boost_played = right_boost_played
            blue_force_played = right_force_played
            blue_levitate_played = right_levitate_played

        if self._debug:
            left_color = (255, 255, 0) if is_flipped else (255, 0, 255)
            right_color = (255, 0, 255) if is_flipped else (255, 255, 0)
            left_boost_box = self._cornersToBox(left_boost_tl, left_boost_br)
            left_force_box = self._cornersToBox(left_force_tl, left_force_br)
            left_levitate_box = self._cornersToBox(left_levitate_tl, left_levitate_br)
            right_boost_box = self._cornersToBox(right_boost_tl, right_boost_br)
            right_force_box = self._cornersToBox(right_force_tl, right_force_br)
            right_levitate_box = self._cornersToBox(right_levitate_tl, right_levitate_br)
            self._drawBox(debug_img, left_boost_box, left_color)
            self._drawBox(debug_img, left_force_box, left_color)
            self._drawBox(debug_img, left_levitate_box, left_color)
            self._drawBox(debug_img, right_boost_box, right_color)
            self._drawBox(debug_img, right_force_box, right_color)
            self._drawBox(debug_img, right_levitate_box, right_color)
            cv2.circle(debug_img, left_boost_loc, 2, left_color, -1)
            cv2.circle(debug_img, left_force_loc, 2, left_color, -1)
            cv2.circle(debug_img, left_levitate_loc, 2, left_color, -1)
            cv2.circle(debug_img, right_boost_loc, 2, right_color, -1)
            cv2.circle(debug_img, right_force_loc, 2, right_color, -1)
            cv2.circle(debug_img, right_levitate_loc, 2, right_color, -1)

        return (
            red_boost_count, red_boost_played,
            red_force_count, red_force_played,
            red_levitate_count, red_levitate_played,
            blue_boost_count, blue_boost_played,
            blue_force_count, blue_force_played,
            blue_levitate_count, blue_levitate_played,
        )

    def _getSwitchScaleInfo(self, img, debug_img, is_flipped):
        left_switch_loc = self._transformPoint((257, 76))
        left_scale_loc = self._transformPoint((257, 51))
        right_switch_loc = self._transformPoint((1279 - 257,  76))
        right_scale_loc = self._transformPoint((1279 - 257, 51))

        left_switch_owned = self._checkSaturated(img, left_switch_loc)
        left_scale_owned = self._checkSaturated(img, left_scale_loc)
        right_switch_owned = self._checkSaturated(img, right_switch_loc)
        right_scale_owned = self._checkSaturated(img, right_scale_loc)

        if is_flipped:
            red_switch_owned = right_switch_owned
            red_scale_owned = right_scale_owned
            blue_switch_owned = left_switch_owned
            blue_scale_owned = left_scale_owned
        else:
            red_switch_owned = left_switch_owned
            red_scale_owned = left_scale_owned
            blue_switch_owned = right_switch_owned
            blue_scale_owned = right_scale_owned

        if self._debug:
            left_color = (255, 255, 0) if is_flipped else (255, 0, 255)
            right_color = (255, 0, 255) if is_flipped else (255, 255, 0)
            cv2.circle(debug_img, left_switch_loc, 2, left_color, -1)
            cv2.circle(debug_img, left_scale_loc, 2, left_color, -1)
            cv2.circle(debug_img, right_switch_loc, 2, right_color, -1)
            cv2.circle(debug_img, right_scale_loc, 2, right_color, -1)

        return (
            red_switch_owned, blue_switch_owned,
            red_scale_owned, blue_scale_owned,
        )

    def _getPowerupInfo(self, img, debug_img):
        # Who owns powerup
        left_point = self._transformPoint((631, 80))
        right_point = self._transformPoint((1279 - 631, 80))
        # Which powerup
        powerup_tl = self._transformPoint((630, 80))
        powerup_br = self._transformPoint((1279 - 630, 105))
        # How much time left
        time_tl = self._transformPoint((624, 50))
        time_br = self._transformPoint((1279 - 624, 79))

        if self._debug:
            time_box = self._cornersToBox(time_tl, time_br)
            self._drawBox(debug_img, time_box, (0, 255, 0))
            cv2.circle(debug_img, left_point, 2, (0, 255, 0), -1)
            cv2.circle(debug_img, right_point, 2, (0, 255, 0), -1)
            powerup_box = self._cornersToBox(powerup_tl, powerup_br)
            self._drawBox(debug_img, powerup_box, (0, 255, 0))

        # Who owns powerup
        left_bgr = img[left_point[1], left_point[0], :]
        right_bgr = img[right_point[1], right_point[0], :]

        if not ((left_bgr[0] > left_bgr[2] and right_bgr[0] > right_bgr[2]) or
                (left_bgr[0] < left_bgr[2] and right_bgr[0] < right_bgr[2])):
            return None, None, None, None

        is_red_powerup = left_bgr[0] < left_bgr[2]  # More red than blue

        # How much time left
        time = self._parseDigits(self._getImgCropThresh(img, time_tl, time_br, white=True))

        # Which powerup
        powerup_img = img[powerup_tl[1]:powerup_br[1], powerup_tl[0]:powerup_br[0]]

        scale = self._transform['scale'] * self._TEMPLATE_SCALE
        best_max_val = 0
        current_powerup = None
        for key, template_img in self._templates.items():
            template_img = cv2.resize(template_img, (int(template_img.shape[0]*scale), int(template_img.shape[1]*scale)))
            res = cv2.matchTemplate(powerup_img, template_img, cv2.TM_CCOEFF)
            _, max_val, _, _ = cv2.minMaxLoc(res)
            if max_val > best_max_val:
                best_max_val = max_val
                current_powerup = key

        if is_red_powerup:
            return (current_powerup, None, time, None)
        else:
            return (None, current_powerup, None, time)

    def _getAutoBoss(self, img, debug_img, is_flipped):
        left_auto_loc = self._transformPoint((550, 54))
        left_boss_loc = self._transformPoint((580, 54))
        right_auto_loc = self._transformPoint((700, 54))
        right_boss_loc = self._transformPoint((730, 54))

        left_auto_sample = img[left_auto_loc[1], left_auto_loc[0], :]
        left_boss_sample = img[left_boss_loc[1], left_boss_loc[0], :]
        right_auto_sample = img[right_auto_loc[1], right_auto_loc[0], :]
        right_boss_sample = img[right_boss_loc[1], right_boss_loc[0], :]

        left_auto_hsv = colorsys.rgb_to_hsv(float(left_auto_sample[2])/255, float(left_auto_sample[1])/255, float(left_auto_sample[0])/255)
        left_boss_hsv = colorsys.rgb_to_hsv(float(left_boss_sample[2])/255, float(left_boss_sample[1])/255, float(left_boss_sample[0])/255)
        right_auto_hsv = colorsys.rgb_to_hsv(float(right_auto_sample[2])/255, float(right_auto_sample[1])/255, float(right_auto_sample[0])/255)
        right_boss_hsv = colorsys.rgb_to_hsv(float(right_boss_sample[2])/255, float(right_boss_sample[1])/255, float(right_boss_sample[0])/255)

        left_auto_quest = left_auto_hsv[0] > 0.116 and left_auto_hsv[0] < 0.216
        left_face_the_boss = left_boss_hsv[0] > 0.116 and left_boss_hsv[0] < 0.216
        right_auto_quest = right_auto_hsv[0] > 0.116 and right_auto_hsv[0] < 0.216
        right_face_the_boss = right_boss_hsv[0] > 0.116 and right_boss_hsv[0] < 0.216

        if is_flipped:
            red_auto_quest = right_auto_quest
            red_face_the_boss = right_face_the_boss
            blue_auto_quest = left_auto_quest
            blue_face_the_boss = left_face_the_boss
        else:
            red_auto_quest = left_auto_quest
            red_face_the_boss = left_face_the_boss
            blue_auto_quest = right_auto_quest
            blue_face_the_boss = right_face_the_boss

        if self._debug:
            left_color = (255, 255, 0) if is_flipped else (255, 0, 255)
            right_color = (255, 0, 255) if is_flipped else (255, 255, 0)
            cv2.circle(debug_img, left_auto_loc, 2, left_color, -1)
            cv2.circle(debug_img, left_boss_loc, 2, left_color, -1)
            cv2.circle(debug_img, right_auto_loc, 2, right_color, -1)
            cv2.circle(debug_img, right_boss_loc, 2, right_color, -1)

        return (
            red_auto_quest, blue_auto_quest,
            red_face_the_boss, blue_face_the_boss,
        )

    def _getMatchDetails(self, img, force_find_overlay):
        debug_img = None
        if self._debug:
            debug_img = img.copy()

        time_remaining, mode = self._getTimeAndMode(img, debug_img)
        if mode in {'pre_match', 'post_match'} or force_find_overlay:
            self._match_key = None
        match_key, match_name = self._getMatchKeyName(img, debug_img)
        is_flipped = self._getFlipped(img, debug_img)
        red_score, blue_score = self._getScores(img, debug_img, is_flipped)

        (
            red_boost_count, red_boost_played,
            red_force_count, red_force_played,
            red_levitate_count, red_levitate_played,
            blue_boost_count, blue_boost_played,
            blue_force_count, blue_force_played,
            blue_levitate_count, blue_levitate_played,
        ) = self._getVaultInfo(img, debug_img, is_flipped)
        (
            red_switch_owned, blue_switch_owned,
            red_scale_owned, blue_scale_owned,
        ) = self._getSwitchScaleInfo(img, debug_img, is_flipped)
        (
            red_current_powerup, blue_current_powerup,
            red_powerup_time_remaining, blue_powerup_time_remaining,
        ) = self._getPowerupInfo(img, debug_img)
        (
            red_auto_quest, blue_auto_quest,
            red_face_the_boss, blue_face_the_boss,
        ) = self._getAutoBoss(img, debug_img, is_flipped)

        if self._debug:
            cv2.imshow("ROIs", debug_img)
            cv2.waitKey(1)

        if match_key is not None and red_score is not None \
                and blue_score is not None and time_remaining is not None:
            return OngoingMatchDetails(
                match_key=match_key,
                match_name=match_name,
                mode=mode,
                time=time_remaining,
                red=Alliance2018(
                    score=red_score,
                    boost_count=red_boost_count,
                    boost_played=red_boost_played,
                    force_count=red_force_count,
                    force_played=red_force_played,
                    levitate_count=red_levitate_count,
                    levitate_played=red_levitate_played,
                    switch_owned=red_switch_owned,
                    scale_owned=red_scale_owned,
                    current_powerup=red_current_powerup,
                    powerup_time_remaining=red_powerup_time_remaining,
                    auto_quest=red_auto_quest,
                    face_the_boss=red_face_the_boss,
                ),
                blue=Alliance2018(
                    score=blue_score,
                    boost_count=blue_boost_count,
                    boost_played=blue_boost_played,
                    force_count=blue_force_count,
                    force_played=blue_force_played,
                    levitate_count=blue_levitate_count,
                    levitate_played=blue_levitate_played,
                    switch_owned=blue_switch_owned,
                    scale_owned=blue_scale_owned,
                    current_powerup=blue_current_powerup,
                    powerup_time_remaining=blue_powerup_time_remaining,
                    auto_quest=blue_auto_quest,
                    face_the_boss=blue_face_the_boss,
                )
            )
        else:
            return None



import numpy


class OverflowPreventer(object):
    """
    A context manager that exposes a numpy array preventing simple operations from overflowing
    Example:
    array= numpy.array( [255], dtype=numpy.uint8 )
    with OverflowPreventer( array ) as prevented:
        prevented+=1
    print array
    """

    inverse_operator = {'__iadd__': '__sub__', '__isub__': '__add__', '__imul__': '__div__', '__idiv__': '__mul__'}
    bypass_operators = ['__str__', '__repr__', '__getitem__']

    def __init__(self, matrix):
        class CustomWrapper(object):
            def __init__(self, matrix):
                assert matrix.dtype == numpy.uint8
                self.overflow_matrix = matrix
                self.overflow_lower_range = float(0)
                self.overflow_upper_range = float(2 ** 8 - 1)
                for op in OverflowPreventer.bypass_operators:
                    setattr(CustomWrapper, op, getattr(self.overflow_matrix, op))

            def _overflow_operator(self, b, forward_operator):
                m, lr, ur = self.overflow_matrix, self.overflow_lower_range, self.overflow_upper_range
                assert type(b) in (int, float)
                reverse_operator = OverflowPreventer.inverse_operator[forward_operator]
                uro = getattr(ur, reverse_operator)
                lro = getattr(lr, reverse_operator)
                afo = getattr(m, forward_operator)
                overflows = m > uro(b)
                underflows = m < lro(b)
                afo(b)
                m[overflows] = ur
                m[underflows] = lr
                return self

            def __getattr__(self, attr):
                if hasattr(self.wrapped, attr):
                    return getattr(self.wrapped, attr)
                else:
                    raise AttributeError

        self.wrapper = CustomWrapper(matrix)
        import functools
        for op in OverflowPreventer.inverse_operator.keys():
            setattr(CustomWrapper, op, functools.partial(self.wrapper._overflow_operator, forward_operator=op))

    def __enter__(self):
        return self.wrapper

    def __exit__(self, type, value, tb):
        pass



def _same_type(a, b):
    type_correct = False
    if type(a) == type(b):
        type_correct = True
    try:
        if isinstance(a, b):
            type_correct = True
    except TypeError:  # v may not be a class or type, but an int, a string, etc
        pass
    return type_correct


def _broadcast(src_processor, src_atr_name, dest_processors, dest_atr_name, transform_function):
    """
    To be used exclusively by create_broadcast.
    A broadcast function gets an attribute on the src_processor and
    sets it (possibly under a different name) on dest_processors
    """
    value = getattr(src_processor, src_atr_name)
    value = transform_function(value)
    for d in dest_processors:
        setattr(d, dest_atr_name, value)


def create_broadcast(src_atr_name, dest_processors, dest_atr_name=None, transform_function=lambda x: x):
    """
    This method creates a function, intended to be called as a
    Processor posthook, that copies some of the processor's attributes
    to other processors
    """
    from functools import partial
    if dest_atr_name == None:
        dest_atr_name = src_atr_name
    if not hasattr(dest_processors, "__iter__"):  # a single processor was given instead
        dest_processors = [dest_processors]
    return partial(_broadcast, src_atr_name=src_atr_name, dest_processors=dest_processors, dest_atr_name=dest_atr_name,
                   transform_function=transform_function)


class Parameters(dict):
    def __add__(self, other):
        d3 = Parameters()
        d3.update(self)
        d3.update(other)
        return d3


class Processor(object):
    """
    In goes something, out goes another. Processor.process() models
    the behaviour of a function, where there are some stored parameters
    in the Processor instance. Further, it optionally calls arbitrary
    functions before and after processing (prehooks, posthooks)
    """

    PARAMETERS = Parameters()

    def __init__(self, **args):
        """sets default parameters"""
        for k, v in self.PARAMETERS.items():
            setattr(self, k, v)
        self.set_parameters(**args)
        self._prehooks = []  # functions (on input) to be executed before processing
        self._poshooks = []  # functions (on output) to be executed after processing

    def get_parameters(self):
        """returns a dictionary with the processor's stored parameters"""
        parameter_names = self.PARAMETERS.keys()
        # TODO: Unresolved reference for processor
        parameter_values = [getattr(processor, n) for n in parameter_names]
        return dict(zip(parameter_names, parameter_values))

    def set_parameters(self, **args):
        """sets the processor stored parameters"""
        for k, v in self.PARAMETERS.items():
            new_value = args.get(k)
            if new_value != None:
                if not _same_type(new_value, v):
                    raise Exception(
                        "On processor {0}, argument {1} takes something like {2}, but {3} was given".format(self, k, v,
                                                                                                            new_value))
                setattr(self, k, new_value)
        not_used = set(args.keys()).difference(set(self.PARAMETERS.keys()))
        not_given = set(self.PARAMETERS.keys()).difference(set(args.keys()))
        return not_used, not_given

    def _process(self, arguments):
        raise NotImplementedError(str(self.__class__) + "." + "_process")

    def add_prehook(self, prehook_function):
        self._prehooks.append(prehook_function)

    def add_poshook(self, poshook_function):
        self._poshooks.append(poshook_function)

    def process(self, arguments):
        self._input = arguments
        for prehook in self._prehooks:
            prehook(self)
        output = self._process(arguments)
        self._output = output
        for poshook in self._poshooks:
            poshook(self)
        return output


class DisplayingProcessor(Processor):
    def display(self, display_before=False):
        """
        Show the last effect this processor had - on a GUI, for
        example. If show_before is True, show the "state before
        processor"
        """
        raise NotImplementedError


class ProcessorStack(Processor):
    """a stack of processors. Each processor's output is fed to the next"""

    def __init__(self, processor_instances=[], **args):
        self.set_processor_stack(processor_instances)
        Processor.__init__(self, **args)

    def set_processor_stack(self, processor_instances):
        assert all(isinstance(x, Processor) for x in processor_instances)
        self.processors = processor_instances

    def get_parameters(self):
        """gets from all wrapped processors"""
        d = {}
        for p in self.processors:
            parameter_names = list(p.PARAMETERS.keys())
            parameter_values = [getattr(p, n) for n in parameter_names]
            d.update(dict(zip(parameter_names, parameter_values)))
        return d

    def set_parameters(self, **args):
        """sets to all wrapped processors"""
        not_used = set()
        not_given = set()
        for p in self.processors:
            nu, ng = p.set_parameters(**args)
            not_used = not_used.union(nu)
            not_given = not_given.union(ng)
        return not_used, not_given

    def _process(self, arguments):
        for p in self.processors:
            arguments = p.process(arguments)
        return arguments


class DisplayingProcessorStack(ProcessorStack):
    def display(self, display_before=False):
        if display_before:
            pr = self.processors[1:]
            self.processors.display(display_before=True)
        else:
            pr = self.processors
        for p in pr:
            if hasattr(p, "display"):
                p.display(display_before=False)



from .opencv_utils import show_image_and_wait_for_key, BrightnessProcessor, draw_segments
from .segmentation_aux import contained_segments_matrix, LineFinder, guess_segments_lines
from .processor import DisplayingProcessor, create_broadcast
import numpy


def create_default_filter_stack():
    stack = [LargeFilter(), SmallFilter(), LargeAreaFilter(), ContainedFilter(), LineFinder(), NearLineFilter()]
    stack[4].add_poshook(create_broadcast("lines_topmiddlebottoms", stack[5]))
    return stack


class Filter(DisplayingProcessor):
    """A filter processes given segments, returning only the desirable ones"""

    PARAMETERS = DisplayingProcessor.PARAMETERS

    def display(self, display_before=False):
        """shows the effect of this filter"""
        try:
            copy = self.image.copy()
        except AttributeError:
            raise Exception("You need to set the Filter.image attribute for displaying")
        copy = BrightnessProcessor(brightness=0.6).process(copy)
        s, g = self._input, self.good_segments_indexes
        draw_segments(copy, s[g], (0, 255, 0))
        draw_segments(copy, s[True ^ g], (0, 0, 255))
        show_image_and_wait_for_key(copy, "segments filtered by " + self.__class__.__name__)

    def _good_segments(self, segments):
        raise NotImplementedError

    def _process(self, segments):
        good = self._good_segments(segments)
        self.good_segments_indexes = good
        segments = segments[good]
        if not len(segments):
            raise Exception("0 segments after filter " + self.__class__.__name__)
        return segments


class LargeFilter(Filter):
    """desirable segments are larger than some width or height"""
    PARAMETERS = Filter.PARAMETERS + {"min_width": 4, "min_height": 8}

    def _good_segments(self, segments):
        good_width = segments[:, 2] >= self.min_width
        good_height = segments[:, 3] >= self.min_height
        return good_width * good_height  # AND


class SmallFilter(Filter):
    """desirable segments are smaller than some width or height"""
    PARAMETERS = Filter.PARAMETERS + {"max_width": 30, "max_height": 50}

    def _good_segments(self, segments):
        good_width = segments[:, 2] <= self.max_width
        good_height = segments[:, 3] <= self.max_height
        return good_width * good_height  # AND


class LargeAreaFilter(Filter):
    """desirable segments' area is larger than some"""
    PARAMETERS = Filter.PARAMETERS + {"min_area": 45}

    def _good_segments(self, segments):
        return (segments[:, 2] * segments[:, 3]) >= self.min_area


class ContainedFilter(Filter):
    """desirable segments are not contained by any other"""

    def _good_segments(self, segments):
        m = contained_segments_matrix(segments)
        return (True ^ numpy.max(m, axis=1))


class NearLineFilter(Filter):
    PARAMETERS = Filter.PARAMETERS + {"nearline_tolerance": 5.0}  # percentage distance stddev
    '''desirable segments have their y near a line'''

    def _good_segments(self, segments):
        lines = guess_segments_lines(segments, self.lines_topmiddlebottoms, nearline_tolerance=self.nearline_tolerance)
        good = lines != -1
        return good



import colorsys
import cv2
import numpy as np
import pickle
import logging
from PIL import Image
import pkg_resources
import pytesseract
import regex

from .simpleocr_utils.segmentation import segments_to_numpy
from .simpleocr_utils.feature_extraction import SimpleFeatureExtractor


class NoOverlayFoundException(Exception):
    pass


QF_BRACKET_ELIM_MAPPING = {
    1: (1, 1),  # (set, match)
    2: (2, 1),
    3: (3, 1),
    4: (4, 1),
    5: (1, 2),
    6: (2, 2),
    7: (3, 2),
    8: (4, 2),
}

SF_BRACKET_ELIM_MAPPING = {
    1: (1, 1),  # (set, match)
    2: (2, 1),
    3: (1, 2),
    4: (2, 2),
}


NUMBER_PATTERN = '0-9ZSO'  # Characters that might get recognized as numbers
def num_pat(r):
    return r.replace('@', NUMBER_PATTERN)


# Possible formats are like:
# Test Match
# Practice X of Y
# Qualification X of Y
# Octofinal X of Y
# Octofinal Tiebreaker X
# Quarterfinal X of Y
# Quarterfinal Tiebreaker X
# Semifinal X of Y
# Semifinal Tiebreaker X
# Final X
# Final Tiebreaker
# Einstein X of Y
# Einstein Final X
# Einstein Final Tiebreaker
MATCH_ID_FORMATS = [
    (regex.compile(num_pat('(Test\s+Match){e<=3}')), 'test', False),
    (regex.compile(num_pat('(Practice){e<=3}\s+([@]+)'), False), 'pm', False),
    (regex.compile(num_pat('(Qualification){e<=3}\s+([@]+)'), False), 'qm', False),
    (regex.compile(num_pat('(Octofinal){e<=3}\s+([@]+)'), False), 'ef', False),
    (regex.compile(num_pat('(Octofinal\s+Tiebreaker){e<=3}\s+([@]+)'), False), 'ef', True),
    (regex.compile(num_pat('(Quarterfinal){e<=3}\s+([@]+)'), False), 'qf', False),
    (regex.compile(num_pat('(Quarterfinal\s+Tiebreaker){e<=3}\s+([@]+)'), False), 'qf', True),
    (regex.compile(num_pat('(Semifinal){e<=3}\s+([@]+)'), False), 'sf', False),
    (regex.compile(num_pat('(Semifinal\s+Tiebreaker){e<=3}\s+([@]+)'), False), 'sf', True),
    (regex.compile(num_pat('(Final){e<=3}\s+([@]+)'), False), 'f', False),
    (regex.compile(num_pat('(Overtime){e<=3}\s+([@]+)'), False), 'overtimef', False),
    (regex.compile(num_pat('(Einstein){e<=3}\s+([@]+)'), False), 'sf', False),
    (regex.compile(num_pat('(Einstein\s+Final){e<=3}\s+([@]+)'), False), 'f', False),
    (regex.compile(num_pat('(Einstein\s+Final\s+Tiebreaker){e<=3}\s+([@]+)'), False), 'f', True),
]


def fix_digits(text):
    return int(text.replace('Z', '2').replace('S', '5').replace('O', '0'))


class LivescoreBase(object):
    def __init__(self, game_year, debug=False, save_training_data=False, training_data=None):
        self._debug = debug
        self._save_training_data = save_training_data

        self._WHITE_LOW = np.array([120, 120, 120])
        self._WHITE_HIGH = np.array([255, 255, 255])

        self._BLACK_LOW = np.array([0, 0, 0])
        self._BLACK_HIGH = np.array([135, 135, 155])

        self._OCR_HEIGHT = 64  # Do all OCR at this size

        self._transform = None  # scale, tx, ty

        # Setup feature detector and matcher
        self._detector = cv2.xfeatures2d.SURF_create()

        FLANN_INDEX_KDTREE = 0
        index_params = dict(algorithm = FLANN_INDEX_KDTREE, trees = 5)
        search_params = dict(checks = 50)
        self._flann = cv2.FlannBasedMatcher(index_params, search_params)

        self._MIN_MATCH_COUNT = 10

        # Compute score overlay keypoints and descriptors
        self._TEMPLATE_SCALE = 0.5  # lower is faster
        template = cv2.imread(
            pkg_resources.resource_filename(__name__, 'templates') + \
            '/score_overlay_{}.png'.format(game_year))
        self._template = cv2.resize(template, (
            np.int32(np.round(template.shape[1] * self._TEMPLATE_SCALE)),
            np.int32(np.round(template.shape[0] * self._TEMPLATE_SCALE))
        ))
        self._kp1, self._des1 = self._detector.detectAndCompute(self._template, None)

        # For saving training data
        self._training_data = {
            'features': np.ndarray((0, 100)),
            'classes': np.ndarray((0, 1)),
        }

        # Train classifier
        if training_data is None:
            with open(pkg_resources.resource_filename(__name__, 'training_data') + '/digits.pkl', "rb") as f:
                training_data = pickle.load(f, encoding='latin1')
        else:
            with open(training_data, "rb") as f:
                training_data = pickle.load(f)
        self._knn = cv2.ml.KNearest_create()
        self._knn.train(training_data['features'].astype(np.float32), cv2.ml.ROW_SAMPLE, training_data['classes'].astype(np.float32))

    def _findScoreOverlay(self, img, force_find_overlay):
        # Does a quick check to see if overlay moved
        # If it has, finds and sets the 2d transform of the overlay in the image
        # Sets the transform to None if the overlay is not found

        if self._transform is not None and not force_find_overlay:
            y = self._transform['ty']
            x = self._transform['tx']
            scale = self._transform['scale']
            overlay = img[
                max(0, int(y)):min(int(y+self._template.shape[0]*scale), img.shape[0] - 1),
                max(0, int(x)):min(int(x+self._template.shape[1]*scale), img.shape[1] - 1),
            ]
            if overlay.shape[0] != 0 and overlay.shape[1] != 0:
                template_img = cv2.resize(self._template, (int(overlay.shape[1]), int(overlay.shape[0])))
                res = cv2.matchTemplate(overlay, template_img, cv2.TM_CCOEFF)
                min_val, _, _, _ = cv2.minMaxLoc(res)
                if min_val > 1000000000:
                    return

        kp2, des2 = self._detector.detectAndCompute(img, None)
        matches = self._flann.knnMatch(self._des1, des2, k=2)

        # Store all the good matches as per Lowe's ratio test
        good = []
        for m, n in matches:
            if m.distance < 0.7 * n.distance:
                good.append(m)

        if len(good) > self._MIN_MATCH_COUNT:
            src_pts = np.float32([ self._kp1[m.queryIdx].pt for m in good ]).reshape(-1, 1, 2)
            dst_pts = np.float32([ kp2[m.trainIdx].pt for m in good ]).reshape(-1, 1, 2)

            t = cv2.estimateRigidTransform(src_pts, dst_pts, False)
            if t is not None:
                self._transform = {
                    'scale': t[0, 0],
                    'tx': t[0, 2],
                    'ty': t[1, 2],
                }
                return

        self._transform = None
        raise NoOverlayFoundException("Not enough matches are found - {}/{}".format(len(good), self._MIN_MATCH_COUNT))

    def _transformPoint(self, point):
        # Transforms a point from template coordinates to image coordinates
        scale = self._transform['scale'] * self._TEMPLATE_SCALE
        tx = self._transform['tx']
        ty = self._transform['ty']
        return np.int32(np.round(point[0] * scale + tx)), np.int32(np.round(point[1] * scale + ty))

    def _cornersToBox(self, tl, br):
        return np.array([
            [tl[0], tl[1]],
            [br[0], tl[1]],
            [br[0], br[1]],
            [tl[0], br[1]]
        ])

    def _getImgCropThresh(self, img, tl, br, white=False):
        # Crop
        img = img[tl[1]:br[1], tl[0]:br[0]]

        # Scale
        scale = float(self._OCR_HEIGHT) / img.shape[0]
        img = cv2.resize(img, (int(img.shape[1] * scale), int(img.shape[0] * scale)))

        # Threshold
        if white:
            return cv2.inRange(img, self._WHITE_LOW, self._WHITE_HIGH)
        else:
            return cv2.inRange(img, self._BLACK_LOW, self._BLACK_HIGH)

    def _parseDigits(self, img):
        # Crop height to digits
        _, contours, _ = cv2.findContours(img, cv2.RETR_LIST, cv2.CHAIN_APPROX_SIMPLE)
        top = img.shape[1]
        bottom = 0
        for cnt in contours:
            if cv2.contourArea(cnt) > 50:
                x, y, w, h = cv2.boundingRect(cnt)
                top = min(top, y)
                bottom = max(bottom, y + h)
        height = bottom - top
        if height <= 0:
            return None
        img = img[top:bottom, :]
        # Scale to uniform height
        scale = float(self._OCR_HEIGHT) / height
        img = cv2.resize(img, (int(img.shape[1] * scale), int(img.shape[0] * scale)))

        # Find bounds for each digit
        _, contours, _ = cv2.findContours(img, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
        digits = []
        for cnt in filter(lambda c: cv2.contourArea(c) > 100, contours):
            segments = segments_to_numpy([cv2.boundingRect(cnt)])
            extractor = SimpleFeatureExtractor(feature_size=10, stretch=False)
            features = extractor.extract(img, segments)
            x,y,w,h = cv2.boundingRect(cnt)

            if self._save_training_data:
                # Construct clean digit image
                if w > self._OCR_HEIGHT:  # Junk, or more than 1 digit
                    continue

                dim = self._OCR_HEIGHT + 5
                digit_img = np.zeros((dim, dim), np.uint8)
                x2 = dim/2 - w/2
                y2 = dim/2 - h/2
                digit_img[y2:y2+h, x2:x2+w] = img[y:y+h, x:x+w]

                config = '--psm 8 -c tessedit_char_whitelist=1234567890'
                string = pytesseract.image_to_string(
                    Image.fromarray(digit_img),
                    config=config).strip()
                if string and string.isdigit():
                    self._training_data['features'] = np.append(self._training_data['features'], features, axis=0)
                    self._training_data['classes'] = np.append(self._training_data['classes'], [[int(string)]], axis=0)
                return None
            else:
                # Perform classification
                if w > self._OCR_HEIGHT:  # More than 1 digit, fall back to Tesseract
                    logging.warning("Falling back to Tesseract!")
                    padded_img = cv2.copyMakeBorder(img[y:y+h, x:x+w], 5, 5, 5, 5, cv2.BORDER_CONSTANT, None, (0, 0, 0))
                    config = '--psm 8 -c tessedit_char_whitelist=1234567890'
                    string = pytesseract.image_to_string(
                        Image.fromarray(padded_img),
                        config=config).strip()
                    if string and string.isdigit():
                        digits.append((string, segments[0, 0]))
                    continue

                # Use KNN
                digit, _, _, _ = self._knn.findNearest(features, k=3)
                digits.append((int(digit), segments[0, 0]))

        fullNumber = ''
        for digit, _ in sorted(digits, key=lambda x: x[1]):
            fullNumber += str(digit)

        if fullNumber != '':
            return int(fullNumber)

    def _getMatchKey(self, raw_match_name):
        for reg, comp_level, tiebreaker in MATCH_ID_FORMATS:
            match = reg.match(raw_match_name)
            if match:
                # TODO: Make API call to TBA to figure out match key
                if comp_level == 'pm':
                    return 'pm{}'.format(fix_digits(match.group(2)))
                elif comp_level == 'qm':
                    return 'qm{}'.format(fix_digits(match.group(2)))
                elif comp_level == 'ef':
                    return 'ef{}'.format(fix_digits(match.group(2)))  # TODO: not correct
                elif comp_level == 'qf':
                    s, m = QF_BRACKET_ELIM_MAPPING[fix_digits(match.group(2))]
                    if tiebreaker:
                        m = 3
                    return 'qf{}m{}'.format(s, m)
                elif comp_level == 'sf':
                    s, m = SF_BRACKET_ELIM_MAPPING[fix_digits(match.group(2))]
                    if tiebreaker:
                        m = 3
                    return 'sf{}m{}'.format(s, m)
                elif comp_level == 'f':
                    return 'f1m{}'.format(fix_digits(match.group(2)))
                elif comp_level == 'overtimef':
                    return 'f1m{}'.format(3+fix_digits(match.group(2)))
                else:
                    return 'test'
        return None

    def _checkSaturated(self, img, point):
        bgr = img[point[1], point[0], :]
        hsv = colorsys.rgb_to_hsv(float(bgr[2])/255, float(bgr[1])/255, float(bgr[0])/255)
        return hsv[1] > 0.2

    def _drawBox(self, img, box, color):
        cv2.polylines(img, [box], True, color, 2, cv2.LINE_AA)

    def read(self, img, force_find_overlay=False):
        img = cv2.resize(img, (1280, 720))
        self._findScoreOverlay(img, force_find_overlay)
        return self._getMatchDetails(img, force_find_overlay)

    def train(self, img, force_find_overlay=False):
        img = cv2.resize(img, (1280, 720))
        self._findScoreOverlay(img, force_find_overlay)
        self._getMatchDetails(img, force_find_overlay)

    def saveTrainingData(self):
        with open('training_data.pkl', 'wb') as output:
            pickle.dump(self._training_data, output, pickle.HIGHEST_PROTOCOL)



import colorsys
import cv2
from PIL import Image
import pytesseract

from .LivescoreBase import LivescoreBase
from .details import Alliance2017, OngoingMatchDetails


class Livescore2017(LivescoreBase):
    def __init__(self, **kwargs) -> object:
        super(Livescore2017, self).__init__(2017, **kwargs)
        self._match_key = None
        self._match_name = None

    def _getMatchKeyName(self, img, debug_img):
        if self._match_key is None:
            tl = self._transformPoint((155, 6))
            br = self._transformPoint((632, 43))

            config = '--psm 7'
            raw_match_name = pytesseract.image_to_string(
                Image.fromarray(self._getImgCropThresh(img, tl, br)), config=config).strip()
            self._match_key = self._getMatchKey(raw_match_name)
            if self._match_key:
                self._match_name = raw_match_name
            else:
                self._match_name = None

            if self._debug:
                box = self._cornersToBox(tl, br)
                self._drawBox(debug_img, box, (0, 255, 0))

        return self._match_key, self._match_name

    def _getTimeAndMode(self, img, debug_img):
        # Find time remaining
        tl = self._transformPoint((617, 55))
        br = self._transformPoint((667, 81))
        time_remaining = self._parseDigits(self._getImgCropThresh(img, tl, br))

        # Determine mode: 'pre_match', 'auto', 'teleop', or 'post_match'
        mode_point = self._transformPoint((497, 70))
        mode_point2 = self._transformPoint((581, 70))
        mode_sample = img[mode_point[1], mode_point[0], :]
        mode_sample2 = img[mode_point2[1], mode_point2[0], :]

        hsv = colorsys.rgb_to_hsv(float(mode_sample[2])/255, float(mode_sample[1])/255, float(mode_sample[0])/255)
        hsv2 = colorsys.rgb_to_hsv(float(mode_sample2[2])/255, float(mode_sample2[1])/255, float(mode_sample2[0])/255)

        if time_remaining is None:
            mode = None
        if time_remaining == 0:
            if hsv[1] > 0.6 and hsv2[1] > 0.6:  # Both saturated
                mode = 'post_match'
            elif hsv[1] > 0.6:  # First saturated
                mode = 'auto'  # End of auton
            else:
                mode = 'pre_match'
        elif time_remaining <= 15 and hsv2[1] < 0.6:
            mode = 'auto'
        else:
            mode = 'teleop'

        if self._debug:
            box = self._cornersToBox(tl, br)
            self._drawBox(debug_img, box, (0, 255, 0))
            cv2.circle(debug_img, mode_point, 2, (0, 255, 0), -1)
            cv2.circle(debug_img, mode_point2, 2, (0, 255, 0), -1)

        return time_remaining, mode

    def _getFlipped(self, img, debug_img):
        # Sample point to determine red/blue side
        color_point = self._transformPoint((496, 95))
        color_sample = img[color_point[1], color_point[0], :]
        is_flipped = color_sample[0] > color_sample[2]  # More blue than red

        if self._debug:
            cv2.circle(debug_img, color_point, 2, (0, 255, 0), -1)

        return is_flipped

    def _getScores(self, img, debug_img, is_flipped):
        # Left score limits
        left_tl = self._transformPoint((496, 90))
        left_br = self._transformPoint((634, 152))
        # Right score limits
        right_tl = self._transformPoint((644, 90))
        right_br = self._transformPoint((784, 152))

        left_score = self._parseDigits(self._getImgCropThresh(img, left_tl, left_br, white=True))
        right_score = self._parseDigits(self._getImgCropThresh(img, right_tl, right_br, white=True))

        if is_flipped:
            red_score = right_score
            blue_score = left_score
        else:
            red_score = left_score
            blue_score = right_score

        if self._debug:
            left_box = self._cornersToBox(left_tl, left_br)
            right_box = self._cornersToBox(right_tl, right_br)
            self._drawBox(debug_img, left_box, (255, 255, 0) if is_flipped else (255, 0, 255))
            self._drawBox(debug_img, right_box, (255, 0, 255) if is_flipped else (255, 255, 0))

        return red_score, blue_score

    def _getFuelScores(self, img, debug_img, is_flipped):
        # Left fuel score
        left_tl = self._transformPoint((316, 123))
        left_br = self._transformPoint((362, 152))
        # Left fuel count
        left_fuel_locs = [
            self._transformPoint((355, 107)),
            self._transformPoint((348, 86)),
            self._transformPoint((332, 70)),
            self._transformPoint((312, 64)),
            self._transformPoint((289, 70)),
            self._transformPoint((273, 86)),
            self._transformPoint((264, 107)),
            self._transformPoint((272, 130)),
            self._transformPoint((289, 146)),
        ]
        # Right fuel score
        right_tl = self._transformPoint((916, 123))
        right_br = self._transformPoint((963, 152))
        # Right fuel count
        right_fuel_locs = [
            self._transformPoint((925, 107)),
            self._transformPoint((930, 86)),
            self._transformPoint((944, 70)),
            self._transformPoint((967, 64)),
            self._transformPoint((991, 70)),
            self._transformPoint((1007, 86)),
            self._transformPoint((1015, 107)),
            self._transformPoint((1007, 130)),
            self._transformPoint((991, 146)),
        ]

        left_fuel_score = self._parseDigits(self._getImgCropThresh(img, left_tl, left_br))
        right_fuel_score = self._parseDigits(self._getImgCropThresh(img, right_tl, right_br))

        left_fuel_count = 0
        for x, y in left_fuel_locs:
            bgr = img[y, x, :]
            hsv = colorsys.rgb_to_hsv(float(bgr[2])/255, float(bgr[1])/255, float(bgr[0])/255)
            if hsv[1] > 0.2:
                left_fuel_count += 1
            else:
                break

        right_fuel_count = 0
        for x, y in right_fuel_locs:
            bgr = img[y, x, :]
            hsv = colorsys.rgb_to_hsv(float(bgr[2])/255, float(bgr[1])/255, float(bgr[0])/255)
            if hsv[1] > 0.2:
                right_fuel_count += 1
            else:
                break

        if is_flipped:
            red_fuel_score = right_fuel_score
            red_fuel_count = right_fuel_count
            blue_fuel_score = left_fuel_score
            blue_fuel_count = left_fuel_count
        else:
            red_fuel_score = left_fuel_score
            red_fuel_count = left_fuel_count
            blue_fuel_score = right_fuel_score
            blue_fuel_count = right_fuel_count

        if self._debug:
            left_color = (255, 255, 0) if is_flipped else (255, 0, 255)
            right_color = (255, 0, 255) if is_flipped else (255, 255, 0)
            left_box = self._cornersToBox(left_tl, left_br)
            right_box = self._cornersToBox(right_tl, right_br)
            self._drawBox(debug_img, left_box, left_color)
            self._drawBox(debug_img, right_box, right_color)
            for point in left_fuel_locs:
                cv2.circle(debug_img, point, 2, left_color, -1)
            for point in right_fuel_locs:
                cv2.circle(debug_img, point, 2, right_color, -1)


        return red_fuel_score, red_fuel_count, blue_fuel_score, blue_fuel_count

    def _getRotors(self, img, debug_img, is_flipped):
        left_tl = self._transformPoint((210, 123))
        left_br = self._transformPoint((230, 148))
        right_tl = self._transformPoint((1048, 123))
        right_br = self._transformPoint((1068, 148))

        left_rotors = self._parseDigits(self._getImgCropThresh(img, left_tl, left_br))
        right_rotors = self._parseDigits(self._getImgCropThresh(img, right_tl, right_br))

        if is_flipped:
            red_rotors = right_rotors
            blue_rotors = left_rotors
        else:
            red_rotors = left_rotors
            blue_rotors = right_rotors

        if self._debug:
            left_color = (255, 255, 0) if is_flipped else (255, 0, 255)
            right_color = (255, 0, 255) if is_flipped else (255, 255, 0)
            left_box = self._cornersToBox(left_tl, left_br)
            right_box = self._cornersToBox(right_tl, right_br)
            self._drawBox(debug_img, left_box, left_color)
            self._drawBox(debug_img, right_box, right_color)

        return red_rotors, blue_rotors

    def _getTouchpads(self, img, debug_img, is_flipped):
        left_tl = self._transformPoint((100, 123))
        left_br = self._transformPoint((120, 148))
        right_tl = self._transformPoint((1158, 123))
        right_br = self._transformPoint((1178, 148))

        left_touchpads = self._parseDigits(self._getImgCropThresh(img, left_tl, left_br))
        right_touchpads = self._parseDigits(self._getImgCropThresh(img, right_tl, right_br))

        if is_flipped:
            red_touchpads = right_touchpads
            blue_touchpads = left_touchpads
        else:
            red_touchpads = left_touchpads
            blue_touchpads = right_touchpads

        if self._debug:
            left_color = (255, 255, 0) if is_flipped else (255, 0, 255)
            right_color = (255, 0, 255) if is_flipped else (255, 255, 0)
            left_box = self._cornersToBox(left_tl, left_br)
            right_box = self._cornersToBox(right_tl, right_br)
            self._drawBox(debug_img, left_box, left_color)
            self._drawBox(debug_img, right_box, right_color)

        return red_touchpads, blue_touchpads

    def _getMatchDetails(self, img, force_find_overlay):
        debug_img = None
        if self._debug:
            debug_img = img.copy()

        time_remaining, mode = self._getTimeAndMode(img, debug_img)
        if mode in {'pre_match', 'post_match'} or force_find_overlay:
            self._match_key = None
        match_key, match_name = self._getMatchKeyName(img, debug_img)

        is_flipped = self._getFlipped(img, debug_img)
        red_score, blue_score = self._getScores(img, debug_img, is_flipped)
        red_fuel_score, red_fuel_count, blue_fuel_score, blue_fuel_count = self._getFuelScores(img, debug_img, is_flipped)
        red_rotors, blue_rotors = self._getRotors(img, debug_img, is_flipped)
        red_touchpads, blue_touchpads = self._getTouchpads(img, debug_img, is_flipped)

        if self._debug:
            cv2.imshow("ROIs", debug_img)
            cv2.waitKey(1)

        if match_key is not None and red_score is not None \
                and blue_score is not None and time_remaining is not None:
            return OngoingMatchDetails(
                match_key=match_key,
                match_name=match_name,
                mode=mode,
                time=time_remaining,
                red=Alliance2017(
                    score=red_score,
                    fuel_score=red_fuel_score,
                    fuel_count=red_fuel_count,
                    rotor_count=red_rotors,
                    touchpad_count=red_touchpads
                ),
                blue=Alliance2017(
                    score=blue_score,
                    fuel_score=blue_fuel_score,
                    fuel_count=blue_fuel_count,
                    rotor_count=blue_rotors,
                    touchpad_count=blue_touchpads
                )
            )
        else:
            return None



from .numpy_utils import OverflowPreventer
from .processor import DisplayingProcessor
import numpy
import cv2


class ImageProcessor(DisplayingProcessor):
    def display(self, display_before=True):
        if display_before:
            show_image_and_wait_for_key(self._input, "before " + self.__class__.__name__)
        show_image_and_wait_for_key(self._output, "after " + self.__class__.__name__)

    def _process(self, image):
        return self._image_processing(image)

    def _image_processing(self, image):
        raise NotImplementedError(str(self.__class__))


class BrightnessProcessor(ImageProcessor):
    """
    changes image brightness.
    A brightness of -1 will make the image all black;
    one of 1 will make the image all white
    """

    PARAMETERS = ImageProcessor.PARAMETERS + {"brightness": 0.0}

    def _image_processing(self, image):
        b = self.brightness
        assert image.dtype == numpy.uint8
        assert -1 <= b <= 1
        image = image.copy()
        with OverflowPreventer(image) as img:
            img += int(b * 256)
        return image


class ContrastProcessor(ImageProcessor):
    """changes image contrast. a scale of 1 will make no changes"""
    PARAMETERS = ImageProcessor.PARAMETERS + {"scale": 1.0, "center": 0.5}

    def _image_processing(self, image):
        assert image.dtype == numpy.uint8
        image = image.copy()
        s, c = self.scale, self.center
        c = int(c * 256)
        with OverflowPreventer(image) as img:
            if s <= 1:
                img *= s
                img += int(c * (1 - s))
            else:
                img -= c * (1 - 1 / s)
                img *= s
        return image


class BlurProcessor(ImageProcessor):
    """changes image contrast. a scale of 1 will make no changes"""
    PARAMETERS = ImageProcessor.PARAMETERS + {"blur_x": 0, "blur_y": 0}

    def _image_processing(self, image):
        assert image.dtype == numpy.uint8
        image = image.copy()
        x, y = self.blur_x, self.blur_y
        if x or y:
            x += (x + 1) % 2  # opencv needs a
            y += (y + 1) % 2  # odd number...
            image = cv2.GaussianBlur(image, (x, y), 0)
        return image


def ask_for_key(return_arrow_keys=True):
    key = 128
    while key > 127:
        key = cv2.waitKey(0)
        if return_arrow_keys:
            if key in (65362, 65364, 65361, 65363):  # up, down, left, right
                return key
        key %= 256
    return key


def background_color(image, numpy_result=True):
    result = numpy.median(numpy.median(image, 0), 0).astype(numpy.int)
    if not numpy_result:
        try:
            result = tuple(map(int, result))
        except TypeError:
            result = (int(result),)
    return result


def show_image_and_wait_for_key(image, name="Image"):
    """
    Shows an image, outputting name. keygroups is a dictionary of keycodes to functions;
    they are executed when the corresponding keycode is pressed
    """

    print("showing", name, "(waiting for input)")
    cv2.imshow('norm', image)
    return ask_for_key()


def draw_segments(image, segments, color=(255, 0, 0), line_width=1):
    """draws segments on image"""
    for segment in segments:
        x, y, w, h = segment
        cv2.rectangle(image, (x, y), (x + w, y + h), color, line_width)


def draw_lines(image, ys, color=(255, 0, 0), line_width=1):
    """draws horizontal lines"""
    for y in ys:
        cv2.line(image, (0, y), (image.shape[1], y), color, line_width)


def draw_classes(image, segments, classes):
    assert len(segments) == len(classes)
    for s, c in zip(segments, classes):
        x, y, w, h = s
        cv2.putText(image, c, (x, y), 0, 0.5, (128, 128, 128))


def get_opencv_version():
    """
    Return the OpenCV version by checking cv2.__version__
    :return: int, 2 or 3
    """
    if cv2.__version__.startswith("3"):
        return 3
    elif cv2.__version__.startswith("2"):
        return 2
    else:
        # This is for the Ubuntu repositories
        # Not the most efficient logical statement, but the most logical to edit in the future
        return 2



import numpy
import cv2
from .segmentation import region_from_segment
from .opencv_utils import background_color

FEATURE_DATATYPE = numpy.float32
# FEATURE_SIZE is defined on the specific feature extractor instance
FEATURE_DIRECTION = 1  # horizontal - a COLUMN feature vector
FEATURES_DIRECTION = 0  # vertical - ROWS of feature vectors


class FeatureExtractor(object):
    """given a list of segments, returns a list of feature vectors"""
    def extract(self, image, segments):
        raise NotImplementedError()


class SimpleFeatureExtractor(FeatureExtractor):
    def __init__(self, feature_size=10, stretch=False):
        self.feature_size = feature_size
        self.stretch = stretch

    def extract(self, image, segments):
        # image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
        fs = self.feature_size
        bg = background_color(image)

        regions = numpy.ndarray(shape=(0, fs), dtype=FEATURE_DATATYPE)
        for segment in segments:
            region = region_from_segment(image, segment)
            if self.stretch:
                region = cv2.resize(region, (fs, fs))
            else:
                x, y, w, h = segment
                proportion = float(min(h, w)) / max(w, h)
                new_size = (fs, int(fs * proportion)) if min(w, h) == h else (int(fs * proportion), fs)
                region = cv2.resize(region, new_size)
                s = region.shape
                newregion = numpy.ndarray((fs, fs), dtype=region.dtype)
                newregion[:, :] = bg
                newregion[:s[0], :s[1]] = region
                region = newregion
            regions = numpy.append(regions, region, axis=0)
        regions.shape = (len(segments), fs ** 2)
        return regions



from .processor import Processor, DisplayingProcessor
from .opencv_utils import draw_lines, show_image_and_wait_for_key
import numpy
import cv2
from functools import reduce


class SegmentOrderer(Processor):
    PARAMETERS = Processor.PARAMETERS + {"max_line_height": 20, "max_line_width": 10000}

    def _process(self, segments):
        """sort segments in read order - left to right, up to down"""
        # sort_f= lambda r: max_line_width*(r[1]/max_line_height)+r[0]
        # segments= sorted(segments, key=sort_f)
        # segments= segments_to_numpy( segments )
        # return segments
        mlh, mlw = self.max_line_height, self.max_line_width
        s = segments.astype(numpy.uint32)  # prevent overflows
        order = mlw * (s[:, 1] // mlh) + s[:, 0]
        sort_order = numpy.argsort(order)
        return segments[sort_order]


class LineFinder(DisplayingProcessor):
    @staticmethod
    def _guess_lines(ys, max_lines=50, confidence_minimum=0.0):
        """guesses and returns text inter-line distance, number of lines, y_position of first line"""
        ys = ys.astype(numpy.float32)
        compactness_list, means_list, diffs, deviations = [], [], [], []
        start_n = 1
        for k in range(start_n, min(len(ys), max_lines)):
            compactness, classified_points, means = cv2.kmeans(data=ys, K=k, bestLabels=None, criteria=(
            cv2.TERM_CRITERIA_EPS | cv2.TERM_CRITERIA_MAX_ITER, 1, 10), attempts=2, flags=cv2.KMEANS_PP_CENTERS)
            means = numpy.sort(means, axis=0)
            means_list.append(means)
            compactness_list.append(compactness)
            if k < 3:
                tmp1 = [1, 2, 500, 550]  # forge data for bad clusters
            else:
                # calculate the center of each cluster. Assuming lines are equally spaced...
                tmp1 = numpy.diff(means, axis=0)  # diff will be equal or very similar
            tmp2 = numpy.std(tmp1) / numpy.mean(means)  # so variance is minimal
            tmp3 = numpy.sum((tmp1 - numpy.mean(tmp1)) ** 2)  # root mean square deviation, more sensitive than std
            diffs.append(tmp1)
            deviations.append(tmp3)

        compactness_list = numpy.diff(
            numpy.log(numpy.array(compactness_list) + 0.01))  # sum small amount to avoid log(0)
        deviations = numpy.array(deviations[1:])
        deviations[0] = numpy.mean(deviations[1:])
        compactness_list = (compactness_list - numpy.mean(compactness_list)) / numpy.std(compactness_list)
        deviations = (deviations - numpy.mean(deviations)) / numpy.std(deviations)
        aglomerated_metric = 0.1 * compactness_list + 0.9 * deviations

        i = numpy.argmin(aglomerated_metric) + 1
        lines = means_list[i]

        # calculate confidence
        betterness = numpy.sort(aglomerated_metric, axis=0)
        confidence = (betterness[1] - betterness[0]) / (betterness[2] - betterness[1])
        if confidence < confidence_minimum:
            raise Exception("low confidence")
        return lines  # still floating points

    def _process(self, segments):
        segment_tops = segments[:, 1]
        segment_bottoms = segment_tops + segments[:, 3]
        tops = self._guess_lines(segment_tops)
        bottoms = self._guess_lines(segment_bottoms)
        if len(tops) != len(bottoms):
            raise Exception("different number of lines")
        middles = (tops + bottoms) / 2
        topbottoms = numpy.sort(numpy.append(tops, bottoms))
        topmiddlebottoms = numpy.sort(reduce(numpy.append, (tops, middles, bottoms)))
        self.lines_tops = tops
        self.lines_bottoms = bottoms
        self.lines_topbottoms = topbottoms
        self.lines_topmiddlebottoms = topmiddlebottoms
        return segments

    def display(self, display_before=False):
        copy = self.image.copy()
        draw_lines(copy, self.lines_tops, (0, 0, 255))
        draw_lines(copy, self.lines_bottoms, (0, 255, 0))
        show_image_and_wait_for_key(copy, "line starts and ends")


def guess_segments_lines(segments, lines, nearline_tolerance=5.0):
    """
    given segments, outputs a array of line numbers, or -1 if it
    doesn't belong to any
    """
    ys = segments[:, 1]
    closeness = numpy.abs(numpy.subtract.outer(ys, lines))  # each row a y, each collumn a distance to each line
    line_of_y = numpy.argmin(closeness, axis=1)
    distance = numpy.min(closeness, axis=1)
    bad = distance > numpy.mean(distance) + nearline_tolerance * numpy.std(distance)
    line_of_y[bad] = -1
    return line_of_y


def contained_segments_matrix(segments):
    """
    givens a n*n matrix m, n=len(segments), in which m[i,j] means
    segments[i] is contained inside segments[j]
    """
    x1, y1 = segments[:, 0], segments[:, 1]
    x2, y2 = x1 + segments[:, 2], y1 + segments[:, 3]
    n = len(segments)

    x1so, x2so, y1so, y2so = list(map(numpy.argsort, (x1, x2, y1, y2)))
    x1soi, x2soi, y1soi, y2soi = list(map(numpy.argsort, (x1so, x2so, y1so, y2so)))  # inverse transformations
    # let rows be x1 and collumns be x2. this array represents where x1<x2
    o1 = numpy.triu(numpy.ones((n, n)), k=1).astype(bool)
    # let rows be x1 and collumns be x2. this array represents where x1>x2
    o2 = numpy.tril(numpy.ones((n, n)), k=0).astype(bool)
    a_inside_b_x = o2[x1soi][:, x1soi] * o1[x2soi][:, x2soi]  # (x1[a]>x1[b] and x2[a]<x2[b])
    a_inside_b_y = o2[y1soi][:, y1soi] * o1[y2soi][:, y2soi]  # (y1[a]>y1[b] and y2[a]<y2[b])
    a_inside_b = a_inside_b_x * a_inside_b_y
    return a_inside_b



from .opencv_utils import show_image_and_wait_for_key, draw_segments, BlurProcessor, get_opencv_version
from .processor import DisplayingProcessor, DisplayingProcessorStack, create_broadcast
from .segmentation_aux import SegmentOrderer
from .segmentation_filters import create_default_filter_stack
import numpy
import cv2

SEGMENT_DATATYPE = numpy.uint16
SEGMENT_SIZE = 4
SEGMENTS_DIRECTION = 0  # vertical axis in numpy


def segments_from_numpy(segments):
    """reverses segments_to_numpy"""
    segments = segments if SEGMENTS_DIRECTION == 0 else segments.tranpose()
    segments = [map(int, s) for s in segments]
    return segments


def segments_to_numpy(segments):
    """given a list of 4-element tuples, transforms it into a numpy array"""
    segments = numpy.array(segments, dtype=SEGMENT_DATATYPE, ndmin=2)  # each segment in a row
    segments = segments if SEGMENTS_DIRECTION == 0 else numpy.transpose(segments)
    return segments


def region_from_segment(image, segment):
    """given a segment (rectangle) and an image, returns it's corresponding subimage"""
    x, y, w, h = segment
    return image[y:y + h, x:x + w]


class RawSegmenter(DisplayingProcessor):
    """A image segmenter. input is image, output is segments"""

    def _segment(self, image):
        """segments an opencv image for OCR. returns list of 4-element tuples (x,y,width, height)."""
        # return segments
        raise NotImplementedError()

    def _process(self, image):
        segments = self._segment(image)
        self.image, self.segments = image, segments
        return segments


class FullSegmenter(DisplayingProcessorStack):
    pass


class RawContourSegmenter(RawSegmenter):
    PARAMETERS = RawSegmenter.PARAMETERS + {"block_size": 11, "c": 10}

    def _segment(self, image):
        self.image = image
        # image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
        image = cv2.adaptiveThreshold(image, maxValue=255, adaptiveMethod=cv2.ADAPTIVE_THRESH_GAUSSIAN_C,
                                      thresholdType=cv2.THRESH_BINARY, blockSize=self.block_size, C=self.c)
        if get_opencv_version() == 3:
            _, contours, hierarchy = cv2.findContours(image, cv2.RETR_LIST, cv2.CHAIN_APPROX_SIMPLE)
        else:
            contours, hierarchy = cv2.findContours(image, cv2.RETR_LIST, cv2.CHAIN_APPROX_SIMPLE)
        segments = segments_to_numpy([cv2.boundingRect(c) for c in contours])
        self.contours, self.hierarchy = contours, hierarchy  # store, may be needed for debugging
        return segments

    def display(self, display_before=False):
        copy = self.image.copy()
        if display_before:
            show_image_and_wait_for_key(copy, "image before segmentation")
        copy.fill(255)
        cv2.drawContours(copy, self.contours, contourIdx=-1, color=(0, 0, 0))
        show_image_and_wait_for_key(copy, "ContourSegmenter contours")
        copy = self.image.copy()
        draw_segments(copy, self.segments)
        show_image_and_wait_for_key(copy, "image after segmentation by " + self.__class__.__name__)


class ContourSegmenter(FullSegmenter):
    def __init__(self, **args):
        filters = create_default_filter_stack()
        stack = [BlurProcessor(), RawContourSegmenter()] + filters + [SegmentOrderer()]
        FullSegmenter.__init__(self, stack, **args)
        stack[0].add_prehook(create_broadcast("_input", filters, "image"))



class Alliance(object):
    def __init__(self, score=None, teams=[]):
        self.teams = teams
        self.score = score


class Alliance2017(Alliance):
    def __init__(
            self,
            score=None,
            teams=[],
            fuel_score=None,
            fuel_count=None,
            rotor_count=None,
            touchpad_count=None):
        super(Alliance2017, self).__init__(score=score, teams=teams)
        self.fuel_score = fuel_score
        self.fuel_count = fuel_count
        self.rotor_count = rotor_count
        self.touchpad_count = touchpad_count

    def getString(self, prefix):
        return (
            '{} score: {}\n'.format(prefix, self.score) + \
            '{} fuel_score: {}\n'.format(prefix, self.fuel_score) + \
            '{} fuel_count: {}\n'.format(prefix, self.fuel_count) + \
            '{} rotor_count: {}\n'.format(prefix, self.rotor_count) + \
            '{} touchpad_count: {}\n'.format(prefix, self.touchpad_count)
        )

class Alliance2018(Alliance):
    def __init__(
            self,
            score=None,
            teams=[],
            boost_count=None,
            boost_played=None,
            force_count=None,
            force_played=None,
            levitate_count=None,
            levitate_played=None,
            switch_owned=None,
            scale_owned=None,
            current_powerup=None,
            powerup_time_remaining=None,
            auto_quest=None,
            face_the_boss=None,
        ):
        super(Alliance2018, self).__init__(score=score, teams=teams)
        self.boost_count = boost_count
        self.boost_played = boost_played
        self.force_count = force_count
        self.force_played = force_played
        self.levitate_count = levitate_count
        self.levitate_played = levitate_played
        self.switch_owned = switch_owned
        self.scale_owned = scale_owned
        self.current_powerup = current_powerup
        self.powerup_time_remaining = powerup_time_remaining
        self.auto_quest = auto_quest
        self.face_the_boss = face_the_boss

    def getString(self, prefix):
        return (
            '{} score: {}\n'.format(prefix, self.score) + \
            '{} force_count: {}\n'.format(prefix, self.force_count) + \
            '{} force_played: {}\n'.format(prefix, self.force_played) + \
            '{} levitate_count: {}\n'.format(prefix, self.levitate_count) + \
            '{} levitate_played: {}\n'.format(prefix, self.levitate_played) + \
            '{} boost_count: {}\n'.format(prefix, self.boost_count) + \
            '{} boost_played: {}\n'.format(prefix, self.boost_played) + \
            '{} switch_owned: {}\n'.format(prefix, self.switch_owned) + \
            '{} scale_owned: {}\n'.format(prefix, self.scale_owned) + \
            '{} current_powerup: {}\n'.format(prefix, self.current_powerup) + \
            '{} powerup_time_remaining: {}\n'.format(prefix, self.powerup_time_remaining) + \
            '{} auto_quest: {}\n'.format(prefix, self.auto_quest) + \
            '{} face_the_boss: {}\n'.format(prefix, self.face_the_boss)
        )

class OngoingMatchDetails(object):
    def __init__(self, match_key=None, match_name=None, mode=None, time=None, red=Alliance(), blue=Alliance()):
        self.match_key = match_key
        self.match_name = match_name
        self.mode = mode
        self.time = time
        self.red = red
        self.blue = blue

    def __str__(self):

        return 'Match Key: {}\nMatch Name: {}\nMode: {}\nTime remaining: {}\n{}{}'.format(
            self.match_key, self.match_name, self.mode, self.time, self.red.getString('Red'), self.blue.getString('Blue'))



from setuptools import setup, find_packages


setup(
    name='livescore',
    version='2018.0.0',
    description='Get FRC scores from an image',
    author='Andrew Dassonville',
    author_email='dassonville.andrew@gmail.com',
    url='https://github.com/andrewda/frc-livescore',
    keywords = ['frc', 'score', 'robotics'],
    license='MIT',
    packages=find_packages(exclude=('tests', 'docs')),
    package_data={'livescore': ['templates/*.png', 'tessdata/*.traineddata', 'training_data/*.pkl']}
)
